# -*- coding: utf-8 -*-
"""
Created on Wed Nov 22 23:26:20 2023

@author: clanglois1
"""
import os

from inspect import getsourcefile
from os.path import abspath

import numpy as np
import tifffile as tf
import time
import cv2
import largestinteriorrectangle as lir

import pyqtgraph as pg
from PyQt5.QtWidgets import QApplication
from PyQt5 import QtGui
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtGui import QPixmap

from inichord import general_functions as gf

from skimage import exposure
from scipy import ndimage as ndi

path2thisFile = abspath(getsourcefile(lambda:0))
uiclass, baseclass = pg.Qt.loadUiType(os.path.dirname(path2thisFile) + "/Series_Stitching_v4.ui")

class MainWindow(uiclass, baseclass):
    def __init__(self,parent):
        super().__init__()
        
        self.setWindowIcon(QtGui.QIcon('icons/Stitch_icon.png'))
        
        self.setupUi(self)
        self.parent = parent

        self.x = 0
        self.y = 0
    
        self.defaultIV() # Hide the PlotWidget until a data has been loaded
        
        # Buttons disabled until first importation
        self.KADBox.setEnabled(False)
        self.col_spinBox.setEnabled(False)
        self.row_spinBox.setEnabled(False)
        self.Range_Edit.setEnabled(False)
        self.Tresh_spinBox.setEnabled(False)
        self.full_range.setEnabled(False)
        self.ComputeClass_bttn.setEnabled(False)
        self.Push_export.setEnabled(False)
        self.Save_bttn.setEnabled(False)
        
        self.OpenData.clicked.connect(self.loaddata) # Open the KAD maps
        self.ComputeClass_bttn.clicked.connect(self.Stitching) # Run the stitching program
        self.Push_export.clicked.connect(self.export_data)
        self.Save_bttn.clicked.connect(self.Save_results) # Save maps and informations
        
        self.KADBox.valueChanged.connect(self.Change_KAD_display)
        self.col_spinBox.valueChanged.connect(self.set_up)
        self.row_spinBox.valueChanged.connect(self.set_up)
        self.flag_KAD = False
        self.flag_affine = True # Affine transformation used as default 
        
        self.val_tresh = self.Tresh_spinBox.value()
        self.Tresh_spinBox.valueChanged.connect(self.Change_treshold)
        
        self.Range_Edit.setText("0")
        self.changeRange() # Modify the searching range 
        self.Range_Edit.editingFinished.connect(self.changeRange) # Modify the searching range 
        
        app = QApplication.instance()
        screen = app.screenAt(self.pos())
        geometry = screen.availableGeometry()
        
        # Control window position and dimensions
        self.move(int(geometry.width() * 0.05), int(geometry.height() * 0.05))
        self.resize(int(geometry.width() * 0.9), int(geometry.height() * 0.6))
        self.screen = screen
        
        # Icons sizes management for QMessageBox
        self.pixmap = QPixmap("icons/Stitch_icon.png")
        self.pixmap = self.pixmap.scaled(100, 100)

#%% Functions
    def data_choice(self): # Allow to apply other treatment depending if the data is a KAD one
        msg = QMessageBox.question(self, 'Image stitching', 'Is it a KAD data ?')
        if msg == QMessageBox.Yes:
            self.flag_KAD = True
        if msg == QMessageBox.No:
            self.flag_KAD = False

    def loaddata(self): 
        self.StackLoc, self.StackDir = gf.getFilePathDialog("Image series (*.tiff)") 
        
        checkimage = tf.TiffFile(self.StackLoc[0]).asarray() # Check for dimension. If 2 dimensions : 2D array. If 3 dimensions : stack of images
        
        if checkimage.ndim != 3: # Check if the data is not a sequence of 3D array
            self.parent.popup_message("3D stitching","Image series (3D stacks) must be imported",'icons/Stitch_icon.png')
            return
        
        else:
            self.Base_map = []
    
            for i in range(0,len(self.StackLoc)):
                Var = tf.TiffFile(self.StackLoc[i]).asarray()
                # Replace NaN value by the nearest OK value
                mask = np.isnan(Var)
                Var[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), Var[~mask])

                if Var.dtype != 'uint8': # Convert to 8bits if needed
                    Var = gf.convertToUint8(Var)
                
                Search0 = np.where(Var == 0.0) # Replace 0.0 by 1 for final crop efficiency
                Var[Search0] = 1
                
                self.Base_map.append(Var)
                
            self.Base_map_display = self.Base_map.copy() 
            
            for i in range(len(self.Base_map_display)): # Flip and rotation for display
                self.Base_map_display[i] = np.flip(self.Base_map_display[i], 1)
                self.Base_map_display[i] = np.rot90(self.Base_map_display[i], k=1, axes=(2, 1))
            
            self.displaySeries(self.Base_map_display[0])
            self.InfoValues_label.setText(str(self.Base_map_display[0].shape))
            
        self.StackLoc, self.StackDir = gf.getFilePathDialog("KAD maps (*.tiff)")  

        checkimage = tf.TiffFile(self.StackLoc[0]).asarray() # Check for dimension. If 2 dimensions : 2D array. If 3 dimensions : stack of images
        
        if checkimage.ndim != 2: # Check if the data is not a sequence of 2D array
            self.parent.popup_message("3D stitching","2D images must be imported",'icons/Stitch_icon.png')
            return
        
        else:
            self.data_choice()
            self.Base_KAD = []
    
            for i in range(0,len(self.StackLoc)):
                Var = tf.TiffFile(self.StackLoc[i]).asarray()
                # Replace NaN value by the nearest OK value
                mask = np.isnan(Var)
                Var[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), Var[~mask])
                # Normalization and CLAHE
                Var = (Var - np.min(Var)) / (np.max(Var) - np.min(Var)) # Normalization step
                Var = exposure.equalize_adapthist(Var, kernel_size=None, clip_limit=0.01, nbins=256) # CLAHE step
    
                if Var.dtype != 'uint8': # Convert to 8bits if needed
                    Var = gf.convertToUint8(Var)
                
                Search0 = np.where(Var == 0.0) # Replace 0.0 by 1 for final crop efficiency
                Var[Search0] = 1
                
                self.Base_KAD.append(Var)
                
            self.Base_KAD_display = self.Base_KAD.copy() 
            
            for i in range(len(self.Base_KAD_display)): # Flip and rotation for display
                self.Base_KAD_display[i] = np.flip(self.Base_KAD_display[i], 0)
                self.Base_KAD_display[i] = np.rot90(self.Base_KAD_display[i], k=1, axes=(1, 0))
            
            self.displayExpKAD(self.Base_KAD_display[0])
    
            # Buttons disabled until first importation
            self.KADBox.setEnabled(True)
            self.col_spinBox.setEnabled(True)
            self.row_spinBox.setEnabled(True)
            self.Range_Edit.setEnabled(True)
            self.Tresh_spinBox.setEnabled(True)
            self.full_range.setEnabled(True)
            self.ComputeClass_bttn.setEnabled(True)
            
        range_step = []
        for i in range(0,len(self.Base_KAD)):
            range_step.append(np.mean(self.Base_KAD[i].shape[1]))

        auto_val_range = int(np.round(np.mean(range_step)*0.3)) # 30% of the mean dimensions of images

        self.Range_Edit.setText(str(auto_val_range))
        self.val_range = auto_val_range
        
    def set_up(self):
        # Here, nbr of column and nbr of raw are defined. self.Super_BSE and self.Super_KAD are created.
        # it allow to work row by row for the next computation step
        self.col_nbr = self.col_spinBox.value()
        self.row_nbr = self.row_spinBox.value()
        self.slice_nbr = len(self.Base_map[0])
        
        self.Super_BSE = []
        self.Super_KAD = []
        
        step = 0
        
        for i in range(0,self.row_nbr):
            self.Super_BSE.append(self.Base_map[step:self.col_nbr + step])
            self.Super_KAD.append(self.Base_KAD[step:self.col_nbr + step])
            
            step = step + self.col_nbr
            
        self.Super_BSE = [x for x in self.Super_BSE if x != []]
        self.Super_KAD = [x for x in self.Super_KAD if x != []]        
    
    def Change_KAD_display(self): # Visualization of the different arrays 
        self.KADBox.setRange(0,len(self.Base_KAD)-1)
        self.KADBox.setSingleStep(1)
        self.Value = self.KADBox.value()
        
        self.displayExpKAD(self.Base_KAD_display[self.Value])
        self.displaySeries(self.Base_map_display[self.Value])
        self.InfoValues_label.setText(str(self.Base_map_display[self.Value].shape))

    def make_transfo_choice(self): # Choice of the transformation to be used to stitch images together
        if self.transfo_choice == 'Translation':
            self.flag_translation = True
            self.flag_affine = False
            self.flag_homo = False
        
        elif self.transfo_choice == 'Affine':
            self.flag_translation = False
            self.flag_affine = True
            self.flag_homo = False
            
        elif self.transfo_choice == 'Homography':
            self.flag_translation = False
            self.flag_affine = False
            self.flag_homo = True

    def Stitching(self): # Stitch images together
        
        self.transfo_choice = self.Transfo_box.currentText() # Extract transformation choice
        self.make_transfo_choice() # Define which transformation must be used
        
        try:
            self.set_up() # Prepare data
            self.transformation_determination() # Check for transformation using the KAD data
            self.apply_transformation() # Apply the transformation on the image series
            
            self.Push_export.setEnabled(True)
            self.Save_bttn.setEnabled(True)
        except:
            self.parent.popup_message("3D stitching","Stitch failed. Please check that the [columns - rows] information match the imported data.",'icons/Stitch_icon.png')
            return

    def transformation_determination(self):
        self.res = []
        self.M_hori = []
        self.M_verti = []
        self.Sel1=[]
        self.Sel2=[]
        
        self.nbr = self.col_nbr * self.row_nbr # Nbr of total 2D array
        
        self.prgbar = 0 # Progress bar initial value
        self.progressBar.setValue(self.prgbar)
        self.progressBar.setRange(0, self.nbr-1)
        self.increment = 0

        if self.col_nbr > 1: # If there are more than 1 column
            for i in range(0,len(self.Super_KAD)):
                for j in range(len(self.Super_KAD[0])-1,0,-1):  
                    # Find matches between images 
                    M = self.find_matches(self.Super_KAD[i][j],self.Super_KAD[i][j-1],direction = "horizontal")
                    # Store transformation matrix M
                    self.M_hori.append(M)
                    # Apply transformation matrix M
                    dst = self.horizontal_stitching(self.Super_KAD[i][j],self.Super_KAD[i][j-1],M)
                    # Stitch the two images together
                    self.Super_KAD[i][j] = self.Super_KAD[i][0:j]
                    self.Super_KAD[i][j-1] = dst
                    
                    QApplication.processEvents()    
                    self.increment = self.increment + 1
                    self.ValSlice = self.increment
                    self.progression_bar()
                    
                    # j is modified because the lenght of Super listing is evolving at each step
                    j = len(self.Super_KAD[i])-1
                    
                    # The twoD stitching is cropped
                    self.cropped_dst,selection = self.cropping_step(dst)
                    self.Sel1.append(selection) 
                    
                # The twoD stitching is stored to be used after 
                self.res.append(self.cropped_dst)

        else : # If there is only 1 column, self.res is a copy of self.KAD_base 
            self.res = self.Base_KAD.copy()

        if self.row_nbr > 1: # If the nbr of row is higher than 1
            for i in range(len(self.res)-1,0,-1):
                # Find matches between images 
                M = self.find_matches(self.res[i],self.res[i-1],direction = "vertical")
                # Store transformation matrix M
                self.M_verti.append(M)
                # Apply transformation matrix M
                dst = self.vertical_stitching(self.res[i],self.res[i-1],M)
                
                # Stitch the two images together
                self.res[i] = self.res[0:i]
                self.res[i-1] = dst
                
                QApplication.processEvents()    
                self.increment = self.increment + 1
                self.ValSlice = self.increment
                self.progression_bar()
                
                # i is modified because the lenght of self.res is evolving at each step
                i = len(self.res[i])-1
                
                # The twoD stitching is cropped
                self.cropped_dst,selection = self.cropping_step(dst)
                self.Sel2.append(selection)
         
        # Apply flip and rotation for display of the stitched 2D arrays
        self.displayed_dst = np.copy(self.cropped_dst)
        self.displayed_dst = np.flip(self.displayed_dst, 0)
        self.displayed_dst = np.rot90(self.displayed_dst, k=1, axes=(1, 0))
                
        self.displayKADStitch(self.displayed_dst)

    def find_matches(self,data1, data2, direction = "horizontal"):
        sift = cv2.SIFT_create()
        
        # Allow to specifiy if the descriptor determination must be apply on the whole images or only a part of it
        if self.full_range.isChecked():
            self.val_range = len(data2[0])

        # create a mask image filled with zeros, the size of original image
        mask = np.zeros(data1.shape[:2], dtype=np.uint8)
        mask2 = np.zeros(data2.shape[:2], dtype=np.uint8)
        
        # Allow to specifiy the range of search to consider the fact that the images are vertical or horizontal
        if direction == 'horizontal':
            cv2.rectangle(mask, (0,0), (self.val_range,len(data1)), (255), thickness = -1)
            cv2.rectangle(mask2, (len(data2[0])-self.val_range,0), (len(data2[0]),len(data2)), (255), thickness = -1)
        if direction == 'vertical':
            cv2.rectangle(mask, (0,0), (len(data1[0]),self.val_range), (255), thickness = -1)
            cv2.rectangle(mask2, (len(data2[0]),len(data2)), (0,len(data2)-self.val_range), (255), thickness = -1)
    
        # Extract descriptor in the two images
        kp1, des1 = sift.detectAndCompute(data1,mask)
        kp2, des2 = sift.detectAndCompute(data2,mask2)
        
        # Look for matches
        match = cv2.BFMatcher()
        matches = match.knnMatch(des1,des2,k=2)
        
        # Matches filtering in order to keep only the best pairs
        good = []
        for m,n in matches:
            if m.distance < self.val_tresh*n.distance:
                good.append(m)
        
        MIN_MATCH_COUNT = 10 # Minimum nbr of descriptor needed to compute
        if len(good) > MIN_MATCH_COUNT:
            src_pts = np.float32([ kp1[m.queryIdx].pt for m in good ]).reshape(-1,1,2)
            dst_pts = np.float32([ kp2[m.trainIdx].pt for m in good ]).reshape(-1,1,2)
        
            # Look for transformaiton matrix between the two set of descriptors
            if self.flag_translation == True : 
                M, mask = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 5.0) #Homography
                
                M[1:3,0] = 0
                M[2,1] = 0
                M[0,1] = 0
                
            elif self.flag_affine == True : 
                M, mask = cv2.estimateAffinePartial2D(src_pts, dst_pts) #Affine
            elif self.flag_homo == True : 
                M, mask = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 5.0) #Homography
        
        else:
            self.parent.popup_message("3D stitching","Not enough descriptor...",'icons/Stitch_icon.png')
            return
            
        return M # Return the matrix of transformation between the 2 images

    def horizontal_stitching(self,data1,data2,M):
        # Specify the initial shape of dst as a function of the transformation (translation/affine: warpAffine ; else: warpPerspective)
        if self.flag_affine == True : 
            dst = cv2.warpAffine(data1,M,(data2.shape[1] + data1.shape[1], data2.shape[0]))
        else :
            dst = cv2.warpPerspective(data1,M,(data2.shape[1] + data1.shape[1], data2.shape[0]))
        
        dst[0:data2.shape[0],0:data2.shape[1]] = data2
        
        return dst
    
    def vertical_stitching(self,data1,data2,M):   
        # Specify the initial shape of dst as a function of the transformation (translation/affine: warpAffine ; else: warpPerspective)
        if self.flag_affine == True : 
            dst = cv2.warpAffine(data1,M,(data2.shape[1] + data1.shape[1], data2.shape[0] + data1.shape[0]))
        else:
            dst = cv2.warpPerspective(data1,M,(data2.shape[1] + data1.shape[1], data2.shape[0] + data1.shape[0]))

        dst[0:data2.shape[0],0:data2.shape[1]] = data2
        
        return dst
        
    def cropping_step(self,dst): # Crop images to remove black border 
        Min_Aligned_stack = self.Mask_min(dst,0)  
        Min_Aligned_stack = Min_Aligned_stack.astype("bool")
        Min_Aligned_stack = ndi.binary_fill_holes(Min_Aligned_stack).astype("bool")
    
        Selection = lir.lir(Min_Aligned_stack) # array([2, 2, 4, 7])
        Cropped_dst = dst[Selection[1]:Selection[3],Selection[0]:Selection[2]]  
        
        return Cropped_dst, Selection
    
    def apply_transformation(self): # Apply the transformation matrices on the 3D array sequence
        self.res_serie = []
        
        self.prgbar = 0 # Progress bar initial value
        self.progressBar.setValue(self.prgbar)
        self.progressBar.setRange(0, self.nbr-1)
        self.increment = 0
        
        for j in range(0,self.slice_nbr):
            k = 0  # Count for apply transformation matrix
            
            self.tempo_res = []
            
            if self.col_nbr > 1: # If there are more than 1 column            
                for a in range(0,len(self.Super_BSE)):
                    for i in range(len(self.Super_BSE[0])-1,0,-1): # 2 then 1
                        if i == len(self.Super_BSE[0])-1:
                            data1 = self.Super_BSE[a][i][j]
                            
                        data2 = self.Super_BSE[a][i-1][j]
                        
                        if self.flag_affine == True : 
                            dst = cv2.warpAffine(data1,self.M_hori[k],(data2.shape[1] + data1.shape[1], data2.shape[0]))
                        else :
                            dst = cv2.warpPerspective(data1,self.M_hori[k],(data2.shape[1] + data1.shape[1], data2.shape[0]))
                        
                        dst[0:data2.shape[0],0:data2.shape[1]] = data2
                        
                        cropped_dst = dst[self.Sel1[k][1]:self.Sel1[k][3],self.Sel1[k][0]:self.Sel1[k][2]]
                        
                        k = k+1 # Count is increased
                        data1 = dst
                        
                        QApplication.processEvents()    
                        self.increment = self.increment + 1
                        self.ValSlice = self.increment
                        self.progression_bar()
                        
                    self.tempo_res.append(cropped_dst)
            else :
                self.X = self.Super_BSE.copy()
                            
                self.tempo_res = []
                for y in range(0,len(self.X)):
                    var = self.X [y][0][j]
                
                    self.tempo_res.append(var)
            
            k = 0 
                
            for i in range(len(self.tempo_res)-1,0,-1):
                if i == len(self.Super_BSE)-1:
                    data1 = self.tempo_res[i]
                    
                data2 = self.tempo_res[i-1]
                
                if self.flag_affine == True : 
                    dst = cv2.warpAffine(data1,self.M_verti[k],(data2.shape[1] + data1.shape[1], data2.shape[0] + data1.shape[0]))
                    dst[0:data2.shape[0],0:data2.shape[1]] = data2
                else : 
                    dst = cv2.warpPerspective(data1,self.M_verti[k],(data2.shape[1] + data1.shape[1], data2.shape[0] + data1.shape[0]))
                    dst[0:data2.shape[0],0:data2.shape[1]] = data2
                
                cropped_dst = dst[self.Sel2[k][1]:self.Sel2[k][3],self.Sel2[k][0]:self.Sel2[k][2]]
                
                k = k+1 # Count is increased
                
                data1 = dst
                
                QApplication.processEvents()    
                self.increment = self.increment + 1
                self.ValSlice = self.increment
                self.progression_bar()
            
            self.res_serie.append(cropped_dst)
        
        self.res_serie_cropped = np.copy(self.res_serie)
        
        self.displayed_res_serie = np.copy(self.res_serie_cropped)
        self.displayed_res_serie = np.flip(self.displayed_res_serie, 1)
        self.displayed_res_serie = np.rot90(self.displayed_res_serie, k=1, axes=(2, 1))
                
        self.displaySeriesStitch(self.displayed_res_serie)
    
    def changeRange(self): # Take into account the searching range modification
        self.val_range = int(self.Range_Edit.text())
        self.full_range.setChecked(False)
        
    def Change_treshold(self): # Take into account the descriptor filtering threshold
        self.val_tresh = self.Tresh_spinBox.value()
        
    def Mask_min(self,Min_Aligned_stack, threshold):
        Mask = np.zeros((len(Min_Aligned_stack),len(Min_Aligned_stack[0])))
        Mask[Min_Aligned_stack > threshold] = 1
        return Mask
    
    def progression_bar(self): # Fonction relative à la barre de progression
        self.prgbar = self.ValSlice
        self.progressBar.setValue(self.prgbar)

    def Save_results(self):
        ti = time.strftime("%Y-%m-%d__%Hh-%Mm-%Ss") # Absolute time 
        
        directory = "Stitched_map_Series_" + ti # Name of the main folder
        PathDir = os.path.join(self.StackDir, directory)  # where to create the main folder
        os.mkdir(PathDir)  # Create main folder

        # Images saving step
        tf.imwrite(PathDir + '/Stitched_map.tiff', np.rot90(np.flip(self.displayed_dst, 0), k=1, axes=(1, 0)).astype('float32')) 
        tf.imwrite(PathDir + '/Stitched_serie.tiff', np.rot90(np.flip(self.displayed_res_serie, 1), k=1, axes=(2, 1)).astype('float32')) 

        # Information (.TXT) step
        with open(PathDir + '\map and serie stitching.txt', 'w') as file:
            file.write("Number of images: " + str(self.slice_nbr))
            file.write("\nNumber of row: " + str(self.row_nbr))   
            file.write("\nNumber of column: " + str(self.col_nbr))  
            if self.full_range.isChecked():
                file.write("\nSearching range (pxls): full images")   
            else:
                file.write("\nSearching range (pxls): "+ str(self.val_range))  
            file.write("\nTransformation: " + str(self.val_tresh)) 
            file.write("\nThreshold: " + str(self.transfo_choice)) 

        # Finished message
        self.parent.popup_message("3D stitching","Saving process is over.",'icons/Stitch_icon.png')

    def export_data(self): # Push stitched image in the main GUI
        # Stitched map (KAD or contour)
        if self.flag_KAD == True: # If stitching has been applied on KAD data 
            self.parent.flag_stitchKAD = True
            
            self.parent.KAD = np.copy(self.displayed_dst) # Copy in the main GUI
            self.parent.StackList.append(self.displayed_dst) # Add the data in the stack list
            
            self.parent.StackDir = self.StackDir
            
            self.parent.displayDataview(self.parent.KAD) # Display the labeled grain
            self.parent.choiceBox.setCurrentIndex(self.parent.choiceBox.count() - 1) # Show the last data in the choiceBox QComboBox
        else :
            self.parent.flag_stitchKAD = False
            
            self.parent.Contour_map = np.copy(self.displayed_dst) # Copy in the main GUI
            self.parent.StackList.append(self.displayed_dst) # Add the data in the stack list
            
            self.parent.StackDir = self.StackDir
            
            self.parent.displayDataview(self.parent.Contour_map) # Display the labeled grain
            self.parent.choiceBox.setCurrentIndex(self.parent.choiceBox.count() - 1) # Show the last data in the choiceBox QComboBox
            
        Combo_text = '\u2022 Stitched map'
        Combo_data = self.displayed_dst
        self.parent.choiceBox.addItem(Combo_text, Combo_data) # Add the data in the QComboBox
     
        self.parent.Info_box.ensureCursorVisible()
        self.parent.Info_box.insertPlainText("\n \u2022 Stitched map.") 
        
        # Stitched serie
        self.parent.flag = False # For auto-denoising access
        self.parent.flag_image = True # For GRDD - GDS
        self.parent.Current_stack = np.copy(self.displayed_res_serie) # Copy in the main GUI
        self.parent.image = np.copy(self.displayed_res_serie) # Copy for the GROD-GOS computation
        self.parent.StackList.append(self.displayed_res_serie) # Add the data in the stack list
        
        self.parent.StackDir = self.StackDir
        
        Combo_text = '\u2022 Stitched serie'
        Combo_data = self.displayed_res_serie
        self.parent.choiceBox.addItem(Combo_text, Combo_data) # Add the data in the QComboBox

        self.parent.displayExpStack(self.parent.Current_stack) # Display the labeled grain
        self.parent.label_Treatment.setText("Stitched serie") # Title of the data
     
        self.parent.Info_box.ensureCursorVisible()
        self.parent.Info_box.insertPlainText("\n \u2022 Stitched serie.")
             
        # Activation of buttons in the main GUI
        self.parent.Edit_tools_button.setEnabled(True)
        self.parent.Registration_button.setEnabled(True)
        self.parent.Background_remover_button.setEnabled(True)
        self.parent.Remove_outliers_button.setEnabled(True)
        self.parent.Manual_denoising_button.setEnabled(True)
        self.parent.Auto_denoising_button.setEnabled(True)
        self.parent.Save_button.setEnabled(True)
        self.parent.Reload_button.setEnabled(True)
        self.parent.Choice_denoiser.setEnabled(True)
        self.parent.choiceBox.setEnabled(True)
        
        for i in range(0,3): # enables [Bleach correction ; STD map ; KAD map]
            self.parent.Tool_choice.model().item(i).setEnabled(True)
        
        # Finished message
        self.parent.popup_message("3D stitching","Stitched KAD has been exported to the main GUI.",'icons/Stitch_icon.png')
        
    def displaySeries(self, series): # Display of initial KAD maps
        self.Series.ui.histogram.hide()
        self.Series.ui.roiBtn.hide()
        self.Series.ui.menuBtn.hide()
        
        view = self.Series.getView()
        state = view.getState()        
        self.Series.setImage(series) 
        view.setState(state)
        
        view.setBackgroundColor(self.parent.color1)
        
        self.Series.autoRange()
        
        ROIplot = self.Series.getRoiPlot()
        ROIplot.setBackground(self.parent.color1)
        
        font=QtGui.QFont('Noto Sans Cond', 8)
        ROIplot.getAxis("bottom").setTextPen('k') # Apply size of the ticks label
        ROIplot.getAxis("bottom").setTickFont(font)
        
        self.Series.timeLine.setPen(color=self.parent.color3, width=15)
        self.Series.frameTicks.setPen(color=self.parent.color1, width=5)
        self.Series.frameTicks.setYRange((0, 1))

        s = self.Series.ui.splitter
        s.handle(1).setEnabled(True)
        s.setStyleSheet("background: 5px white;")
        s.setHandleWidth(5) 
    
    def displayExpKAD(self, series): # Display of initial KAD maps
        self.KADSeries.ui.histogram.hide()
        self.KADSeries.ui.roiBtn.hide()
        self.KADSeries.ui.menuBtn.hide()
        
        view = self.KADSeries.getView()
        state = view.getState()        
        self.KADSeries.setImage(series) 
        view.setState(state)
        view.setBackgroundColor(self.parent.color1)
        
        self.KADSeries.autoRange()
        
    def displayKADStitch(self, series): # Display of initial KAD map
        self.StitchKAD.ui.histogram.show()
        self.StitchKAD.ui.roiBtn.hide()
        self.StitchKAD.ui.menuBtn.hide()
        
        view = self.StitchKAD.getView()
        state = view.getState()        
        self.StitchKAD.setImage(series) 
        view.setState(state)
        view.setBackgroundColor(self.parent.color1)
        
        self.StitchKAD.autoRange()
        
        histplot = self.StitchKAD.getHistogramWidget()
        histplot.setBackground(self.parent.color1)
        
        histplot.region.setBrush(pg.mkBrush(self.parent.color5 + (120,)))
        histplot.region.setHoverBrush(pg.mkBrush(self.parent.color5 + (60,)))
        histplot.region.pen = pg.mkPen(self.parent.color5)
        histplot.region.lines[0].setPen(pg.mkPen(self.parent.color5, width=2))
        histplot.region.lines[1].setPen(pg.mkPen(self.parent.color5, width=2))
        histplot.fillHistogram(color = self.parent.color5)        
        histplot.autoHistogramRange()       
        
    def displaySeriesStitch(self, series): # Display of initial KAD map
        self.StitchSeries.ui.histogram.show()
        self.StitchSeries.ui.roiBtn.hide()
        self.StitchSeries.ui.menuBtn.hide()
        
        view = self.StitchSeries.getView()
        state = view.getState()        
        self.StitchSeries.setImage(series) 
        view.setState(state)
        view.setBackgroundColor(self.parent.color1)
        
        self.StitchSeries.autoRange()
        
        histplot = self.StitchSeries.getHistogramWidget()
        histplot.setBackground(self.parent.color1)
        
        histplot.region.setBrush(pg.mkBrush(self.parent.color5 + (120,)))
        histplot.region.setHoverBrush(pg.mkBrush(self.parent.color5 + (60,)))
        histplot.region.pen = pg.mkPen(self.parent.color5)
        histplot.region.lines[0].setPen(pg.mkPen(self.parent.color5, width=2))
        histplot.region.lines[1].setPen(pg.mkPen(self.parent.color5, width=2))
        histplot.fillHistogram(color = self.parent.color5)        
        histplot.autoHistogramRange()    
        
        ROIplot = self.StitchSeries.getRoiPlot()
        ROIplot.setBackground(self.parent.color1)
        
        font=QtGui.QFont('Noto Sans Cond', 8)
        ROIplot.getAxis("bottom").setTextPen('k') # Apply size of the ticks label
        ROIplot.getAxis("bottom").setTickFont(font)
        
        self.StitchSeries.timeLine.setPen(color=self.parent.color3, width=15)
        self.StitchSeries.frameTicks.setPen(color=self.parent.color1, width=5)
        self.StitchSeries.frameTicks.setYRange((0, 1))

        s = self.StitchSeries.ui.splitter
        s.handle(1).setEnabled(True)
        s.setStyleSheet("background: 5px white;")
        s.setHandleWidth(5) 

    def defaultIV(self):
        self.Series.ui.histogram.hide()
        self.Series.ui.roiBtn.hide()
        self.Series.ui.menuBtn.hide()
        
        view = self.Series.getView()
        view.setBackgroundColor(self.parent.color1)
        
        ROIplot = self.Series.getRoiPlot()
        ROIplot.setBackground(self.parent.color1)
        
        self.KADSeries.ui.histogram.hide()
        self.KADSeries.ui.roiBtn.hide()
        self.KADSeries.ui.menuBtn.hide()
        
        view = self.KADSeries.getView()
        view.setBackgroundColor(self.parent.color1)
        
        ROIplot = self.KADSeries.getRoiPlot()
        ROIplot.setBackground(self.parent.color1)
        
        self.StitchKAD.ui.histogram.hide()
        self.StitchKAD.ui.roiBtn.hide()
        self.StitchKAD.ui.menuBtn.hide()
        
        view = self.StitchKAD.getView()
        view.setBackgroundColor(self.parent.color1)
        
        self.StitchSeries.ui.histogram.hide()
        self.StitchSeries.ui.roiBtn.hide()
        self.StitchSeries.ui.menuBtn.hide()
        
        view = self.StitchSeries.getView()
        view.setBackgroundColor(self.parent.color1)
        
        ROIplot = self.StitchSeries.getRoiPlot()
        ROIplot.setBackground(self.parent.color1)


# -*- coding: utf-8 -*-
"""
Fonctions générales
"""
import os
import h5py
import numpy as np
import math
import tkinter as tk
from tkinter import filedialog
# import cupy as cp
import time
from skimage.morphology import disk, ball
import scipy.ndimage as sci
from skimage.restoration import denoise_nl_means
from skimage import img_as_float
import cv2

import sklearn.preprocessing as pp
import matplotlib.pyplot as plt
import bm3d
from pyvsnr import vsnr2d

import tifffile as tf

#______________________________________________________
# h5py functions
def get_dataset_keys(f):
    keys = []
    f.visit(lambda key : keys.append(key) if isinstance(f[key], h5py.Dataset) else None)
    return keys

def get_group_keys(f):
    keys = []
    f.visit(lambda key : keys.append(key) if isinstance(f[key], h5py.Group) else None)
    return keys

def openH5file(text = 'theoretical database (*.crddb)'):

    '''
    Open a dialog to look for the H5 file to open
    
    Returns (in this order):
        - H5file itself, to be used to get datasets
        - directory of the file
        - file path
        - the list of keys in the H5 file
    '''
   
    filePath, dirFile = getFilePathDialog(text)

    # lecture du fichier de profils theoriques
    f = h5py.File(filePath[0], 'r')

    return f, dirFile, filePath[0], get_dataset_keys(f)

def Saving_img_or_stack(stack):
    """
    Save an image or a stack of images in tif format,
    in its original type (uint8 or float32...)

    Parameters
    ----------
    stack : numpy array 2D or 3D or color image
        DESCRIPTION.

    Returns nothing
    -------
    None.

    """
    
    SavePath = filedialog.asksaveasfilename(title="Save the aligned stack", filetypes=[("TIFF Files", "*.tif")])

    tf.imwrite(SavePath, stack)    
    
def savingMTEXfromHDF5():
    # Ouverture du fichier h5
    
    f, dirFile, filePath, listKeys = openH5file()
    
    
    for i in listKeys:
        if "nScoresOri" in i:
            nScoresOri = np.asarray(f[i])
            
    Quat=nScoresOri[-1,:,:,:]
    x = len(Quat[0])
    y = len(Quat[0][0])

    ti = time.strftime("%Y-%m-%d__%Hh-%Mm")

    with open(dirFile + '\indexGPU_'+ ti + '.quatCHORDv3-CTFxyConv.txt', 'w') as file:
    
        for i in range(x):
            for j in range(y):

                index = 1    
                if Quat[0,i,j] ==0 :
                    index = 0
                file.write(str(index) + '\t' + str(j) + '\t' + str(i) + '\t' + str(Quat[0, i, j]) +
                '\t' + str(Quat[1, i, j])  + '\t' + str(Quat[2, i, j])  + '\t' + str(Quat[3, i, j]) + '\n') 

#________________________________________________________
# type conversion

def scale_to(x, x_min, x_max, t_min, t_max): # Conversion en 16bits
    """
    Scales x to lie between t_min and t_max
    Links:
         https://stats.stackexchange.com/questions/281162/scale-a-number-between-a-range
         https://stats.stackexchange.com/questions/178626/how-to-normalize-data-between-1-and-1
    """
    r = x_max - x_min
    r_t = t_max - t_min
    assert(math.isclose(0,r, abs_tol=np.finfo(float).eps) == False)
    x_s = r_t * (x - x_min) / r + t_min
    return x_s

#________________________________________________________
# GPU function with cupy

def isCupy(a):
    '''
    Parameters
    ----------
    a : any python object
        the object which type to be determined, cupy or not cupy.

    Returns
    -------
    Boolean
        True if the object in parameter is a cupy object, False for any other type.

    '''
    return str(type(a)) == "<class 'cupy.ndarray'>"

#________________________________________________________
# general functions

def getFilePathDialog(text):
    '''
    
    Parameters
    ----------
    text : Ttype 'string'
        the text to be displayed to describe what type of file should be opened by the dialog

    Returns
    -------
    filePath
        the full path to the selected file

    '''
    window = tk.Tk()
    window.wm_attributes('-topmost', 1)
    window.withdraw()

    # Pour utiliser le programme de traitement stack
    filePath = filedialog.askopenfilenames(title=text)
    dirFile = os.path.dirname(filePath[0])
    
    # Pour indexation ces lignes fonctionne (jusque V16 d'indexation sur GPU)
    # filePath = filedialog.askopenfilename(title=text, multiple=True, parent=window)[0]
    # dirFile = os.path.dirname(filePath)
    
    return filePath, dirFile


#______________________________
# image treatment

def remove_outliers(image, radius=2, threshold=50):
    footprint_function = disk if image.ndim == 2 else ball
    footprint = footprint_function(radius=radius)
    median_filtered = sci.median_filter(image, footprint=footprint)
    outliers = (
        (image > median_filtered + threshold)
        | (image < median_filtered - threshold)
    )
    output = np.where(outliers, median_filtered, image)
    return outliers, output

def NonLocalMeanDenoising(image, param_h = 0.06, fastAlgo = True, size = 5, distance = 6):

    denoisedImage = denoise_nl_means(image, h= param_h, fast_mode = True,
                                     patch_size = size,
                                     patch_distance = distance)
    return denoisedImage

def BM3D(image, psd, isFast):
    if isFast:
        denoisedImage = bm3d.bm3d(image, sigma_psd = psd, stage_arg=bm3d.BM3DStages.HARD_THRESHOLDING)
    else:
        denoisedImage = bm3d.bm3d(image, sigma_psd = psd, stage_arg=bm3d.BM3DStages.ALL_STAGES)

    return denoisedImage

def VSNR_funct(image, filter_):
    
    denoisedImage = vsnr2d(image, filter_)
    
    return denoisedImage

def convertToUint8(stack):
      mn = stack.min()
      mx = stack.max()
    
      mx -= mn
    
      stack = ((stack - mn)/mx) * 255.0
      return np.round(stack).astype(np.uint8)
  
def convertToUfloat32(stack):
      mn = stack.min()
      mx = stack.max()
    
      mx -= mn
    
      stack = ((stack - mn)/mx) * 2_000_000_000
      return np.round(stack).astype(np.float32)
   
def sliceinfo(image):
    mn = image.min()
    mx = image.max()
    print(f"\n _________________\nle type de donnée est {type(mn)} \n le min est {mn} \n et le max est {mx}")

def bleach_ratio(stack):
    Mean_intensity = np.zeros(len(stack))
    Ratio = np.zeros(len(stack))
    Bleached_stack = np.zeros((len(stack),len(stack[0]),len(stack[0][0])))

    for i in range(0,len(stack)):
        Mean_intensity[i] = np.mean(stack[i,:,:])
        Ratio[i] = Mean_intensity[i]/Mean_intensity[0]   
        
        if Ratio[i]<= 1 :  
            Bleached_stack[i,:,:] = stack[i,:,:]*Ratio[i]
        else : 
            Bleached_stack[i,:,:] = stack[i,:,:]/Ratio[i]
    
    return Bleached_stack

def cropped_img_or_stack(data,Pj1,Pi1,Pj2,Pi2):
    if len(np.shape(data)) == 3:
        Cropped_data = data[:,Pi1:Pi2,Pj1:Pj2]
    else :
        Cropped_data = data[Pi1:Pi2,Pj1:Pj2]
    
    return Cropped_data

#______________________________
# Denoising characterization

def Pixels_transformation(arr1, arr2): #Permet de connaitre le nombre de pixel ayant été modifiés
    diff = arr2-arr1 #Différences des deux cartes
    
    Unmodified_pixels_number = np.count_nonzero(diff) # Nombre de pixels n'ayant pas été modifiés
    Total_pixels_number = len(diff)*len(diff[0]) # Nombre total de pixels
    Modified_pixels_proportion = Unmodified_pixels_number * 100 / Total_pixels_number
    
    Var = diff
    Modified_pixels_intensity = Var
    del(Var)
    
    Modified_pixels_intensity[Modified_pixels_intensity == 0] = np.nan
    mean_modified_pixels_intensity = np.nanmean(abs(Modified_pixels_intensity), axis=None)
    
    max_intensity_arr = np.max(arr1)
    mean_proportion_intensity = mean_modified_pixels_intensity * 100 / max_intensity_arr
    
    return Modified_pixels_proportion, mean_modified_pixels_intensity, mean_proportion_intensity

# Resize_stack : pour réduire le nombre de pixels d'une image 
def Binning_stack(Raw_Stack, bin_value = 2, inter = cv2.INTER_CUBIC):
    """
    

    Parameters
    ----------
    Raw_Stack : numpy array
        DESCRIPTION: profiles in dimension, then image height in dimension 1 and image width in dimension 2
    bin_value : TYPE, optional
        DESCRIPTION. The default is 2.
    inter : TYPE, optional
        DESCRIPTION. Type of binning interpolation. The default is cv2.INTER_CUBIC.

    Returns
    -------
    Resize_stk : numpy array
        DESCRIPTION: 

    """
    bin_value = bin_value
    X_binned = int(np.round((len(Raw_Stack[0][0])/bin_value)))
    Y_binned = int(np.round((len(Raw_Stack[0])/bin_value)))
    
    Resize_stk = []              
            
    for i in range(0,len(Raw_Stack)):
        Var = Raw_Stack[i,:,:]
        Resized_img = cv2.resize(Var, dsize=(X_binned, Y_binned), interpolation=inter)
        Resize_stk.append(Resized_img)
    
    Resize_stk = np.dstack(Resize_stk)
    Resize_stk = np.swapaxes(Resize_stk, 2, 0)
    Resize_stk = np.rot90(Resize_stk, k=1, axes=(1, 2))
    Resize_stk = np.fliplr(Resize_stk)
    
    return Resize_stk

#_______________________________
# Grain boundaries highlightment 

def canny_edge_determination(img, threshold1, threshold2): # Permet d'obtenir les contours pour une image
    
    img_blur = cv2.GaussianBlur(img, (3,3), 0) # Blur the image for better edge detection
    edge = cv2.Canny(image=img_blur, threshold1=threshold1, threshold2=threshold2) # Canny Edge Detection
    
    return edge

def Sum_canny_edge_determination(stack, threshold1, threshold2): #Permet d'obtenir les contour d'une stack
    edges = np.zeros((len(stack),len(stack[0]),len(stack[0][0])))
    edges_stack = convertToUint8(stack)

    # Determination des contours sur chaque image 
    for i in range(0,len(edges_stack)):
        edges[i,:,:] = canny_edge_determination(edges_stack[i,:,:],threshold1,threshold2) #Stack, threshold 1, threshold 2
    
    # Sommation des contours de chaque slices
    Sum_edges = edges == 255
    Sum_edges = np.sum(Sum_edges,0)
    
    return edges, Sum_edges

def Sum_canny_edge_proportion_determination(Sum_edges, stack, seuil_JDG): #Permet de resortir un filtre de JDG et un taux de pixels appartement à un JDG

    # Seuillage de "Sum_edges" et calcul du taux de l'image étant une interface
    seuil_JDG = seuil_JDG
    thresh_sum_edges = Sum_edges > seuil_JDG
    edges_proportion = np.round(thresh_sum_edges.sum() * 100 / (len(stack[0])*len(stack[0][0])))

    return thresh_sum_edges, edges_proportion


import os
import sys
import numpy as np
from scipy import ndimage
from scipy import fftpack, signal
from inichord import general_functions as gf
import scipy.io.wavfile
import scipy.signal
# import cupy as cp

def cyclic(a, axProf):
    '''
    a : TYPE numpy array 2D or Cupy array 2D
        DESCRIPTION : les profils sont rangés en ligne ou en colonne
    axProf : Integer : décrit selon quel axe les profils s'étendent.
    Si une ligne représente un profil, alors il faut mettre axe 1
    Si une colonne représente un profil, alors il faut mettre axe 0
        
    return : un tableau où on a copié la première donnée de chaqueprofil à la fin de chaque profil,
             et donc les profils ont une dimension augmentée de 1
    '''
    if axProf == 1:
        i = 0
        n = len(a[:, 0])
        b = a[:, 0].reshape((n, -1))
        
    elif axProf == 0:
        i = 1
        n = len(a[0, :])
        b = a[0, :].reshape((-1, n))        
    else:
        print("le tableau d'entrée doit être en 2D et l'axe = 0 ou 1")
        quit()
    
    arr = np.concatenate((a, b), axis = axProf)
    return arr

def fft_profil(arr2, axProf): # Useless 
    Var_fft = np.fft.rfft(arr2, axis = axProf) #FFT

    return Var_fft 

def derivee(a, nbderiv, ax = 1):
    return np.diff(a, nbderiv, axis = ax)

def reshapeProfilesInLine(profiles, size):
    lines = int(profiles.size / size)
    return profiles.reshape(lines, size)

def Butterworth(arr2, axProf, pass_filter, N_order, Wn_order):
    b, a = scipy.signal.butter(N_order, Wn_order, pass_filter)
    arrHP = scipy.signal.filtfilt(b, a, arr2, axProf)
    
    return arrHP

def downSampleProfiles(profiles, step):
    '''
        Parameters
    ----------
    profile : TYPE numpy array
        DESCRIPTION. c'est le tableau contenant les profils en ligne
    step : TYPE int
        DESCRIPTION c'est le pas de réduction (2, 3, 4...)

    Returns une matrice de profils de longueur réduite rangés en ligne
    -------
    None.

    '''
    if profiles.ndim == 1:
        try:
            profiles = profiles[::step]
            return profiles
        except:
            print("step doit être un entier et diviseur de len(profiles) ; retour des profils intacts")
            return profiles
    
    elif profiles.ndim == 2:
        try:
            profiles = profiles[:, ::step]
            return profiles
        except:
            print("step doit être un entier et diviseur de len(profiles) ; retour des profils intacts")
            return profiles 

def normMatProfiles(matrix, ax = 1):
    norm = np.linalg.norm(matrix, axis = ax, keepdims=True)
    return matrix / norm

def centeredEuclidianNorm(prof, ax = 1):
    '''
    Cette fonction prend un vecteur et retourne un vecteur normalisé centré.
    Signification : on enlève la moyenne et après on le divise par sa norme euclidienne

    Parameters
    ----------
    prof : TYPE numpy array OR cupy array
        DESCRIPTION. vecteur à normaliser-centrer
        Note : la fonction peut prendre une assemblée de profils, d'où l'argument
        ax pour spécifier la dimension le long de laquelle les différents profils sont alignés

    Returns
    -------
    prof : TYPE numpy array
        DESCRIPTION. vecteur normalisé

    '''
           
    if not gf.isCupy(prof):
        mean =np.mean(prof, axis = ax, keepdims=True)
        prof = prof - mean
        del mean
        norm = np.linalg.norm(prof, axis = ax, keepdims=True)
    else:
        mean = cp.mean(prof, axis = ax, keepdims=True)
        prof = prof - mean
        del mean
        norm = cp.linalg.norm(prof, axis = ax, keepdims=True)
   
    
    prof = prof / norm
    return prof

def centeredSTD(prof, ax = 1):
    '''
    Cette fonction prend un vecteur et retourne un vecteur centré et divisé par son écart-type.
    Signification : on enlève la moyenne et après on le divise par l'écart-type'
    Parameters
    ----------
    prof : TYPE numpy array OR cupy array
        DESCRIPTION. vecteur à normaliser-centrer
        Note : la fonction peut prendre une assemblée de profils, d'où l'argument
        ax pour spécifier la dimension le long de laquelle les différents profils sont alignés

    Returns
    -------
    prof : TYPE numpy array
        DESCRIPTION. vecteur normalisé

    '''
           
    if not gf.isCupy(prof):
        mean =np.mean(prof, axis = ax, keepdims=True)
        prof = prof - mean
        del mean
        std = np.std(prof, axis = ax, keepdims=True)
    else:
        mean = cp.mean(prof, axis = ax, keepdims=True)
        prof = prof - mean
        del mean
        std = cp.std(prof, axis = ax, keepdims=True)
   
    prof = prof / std
    return prof

def Profile_modifier(array2D, Workflow, normType, axProf):
    '''
    
    Parameters
    ----------
    array2D : TYPE numpy array 2D or Cupy array 2D
        DESCRIPTION : les profils sont rangés en ligne ou en colonne
    ax : TYPE integer
        DESCRIPTION : décrit selon quel axe les profils s'étendent.
        Si une ligne représente un profil, alors il faut mettre axe 1
        Si une colonne représente un profil, alors il faut mettre axe 0

    Returns un tableau de même type avec des profils de dimension n-1, 
        rangés de la même manière et normalisés en euclidien centré
    -------
    None.

    '''
    if normType == "centered euclidian":
        array2D = centeredEuclidianNorm(array2D, ax = axProf)
    elif normType == "centered_std":
        array2D = centeredSTD(array2D, ax = axProf)  
    
    for i in Workflow:
        if i[0] == 'Diff':
            if i[1] != 0:
                array2D = cyclic(array2D, axProf)
                array2D = derivee(array2D, i[1], ax = axProf)
                array2D = centeredEuclidianNorm(array2D, ax = axProf)
        elif i[0] == 'FFT':
            array2D = fft_profil(array2D, axProf)
            # array2D = centeredEuclidianNorm(array2D, ax = axProf)
        elif i[0] == 'Butterworth':
            array2D = Butterworth(array2D, axProf, i[1], i[2], i[3])
            array2D = centeredEuclidianNorm(array2D, ax = axProf)
            
    return array2D 


def FFT_filtering(stack, freq = 1):
    # stack = stack.asarray().astype(np.float64)
    
    freq = int(freq)
    
    h = len(stack[0])
    w = len(stack[0, 1])
    stack_reshape = np.moveaxis(stack, 0, 2)

    img_number = len(stack[:,])
    img_step = int(360/len(stack[:,]))
    stack_filtered = np.zeros((h, w, img_number))
    # stack_background = np.zeros((h, w, img_number))

    for x in range(h):
        for y in range(w):
            sig  = stack_reshape[x,y]
            sample_freq = fftpack.fftfreq(sig.size, d=img_step/360)
            sig_fft = fftpack.fft(sig)

            low_freq_fft = sig_fft.copy()
            high_freq_fft = sig_fft.copy()

            low_freq_fft[np.abs(sample_freq) <= freq] = 0
            high_freq_fft[np.abs(sample_freq) > freq] = 0

            signal_filtered = fftpack.ifft(low_freq_fft)
            # signal_background = fftpack.ifft(high_freq_fft)

            stack_filtered[x,y] = np.real(signal_filtered)
            # stack_background[x,y] = np.real(signal_background)


    stack_filtered = np.moveaxis(stack_filtered, -1, 0)
    stack_filtered = np.asarray(stack_filtered, dtype=np.float32)

    # stack_background = np.moveaxis(stack_background, -1, 0)
    # stack_background= np.asarray(stack_background, dtype=np.float32)
    
    # return stack_filtered,stack_background
    return stack_filtered

def FFT_filtering_prf_InLines(array2D, freq = 1):

    freq = int(freq)
    
    h = len(array2D)
    w = len(array2D[0])
    print('Nbr de point : ' + str(h))
    print('Nbr de profil : ' + str(w))
    # stack_reshape = np.moveaxis(array2D, 0, 2)

    img_number = len(array2D[0])
    print('img number is :' + str(img_number))
    img_step = int(360/img_number)
    print('step is :' + str(img_step))
    stack_filtered = np.zeros((h, w))
    # stack_background = np.zeros((h, w, img_number))

    for x in range(w):
        sig  = array2D[:,x]
        sample_freq = fftpack.fftfreq(sig.size, d=img_step/360)
        # print(sample_freq)
        sig_fft = fftpack.fft(sig)

        low_freq_fft = sig_fft.copy()
        high_freq_fft = sig_fft.copy()

        low_freq_fft[np.abs(sample_freq) <= freq] = 0
        high_freq_fft[np.abs(sample_freq) > freq] = 0

        signal_filtered = fftpack.ifft(low_freq_fft)
        stack_filtered[:,x] = np.real(signal_filtered)

    stack_filtered = np.asarray(stack_filtered, dtype=np.float32)

    return stack_filtered